/*
** StopWatch.c
** This source code was written by Snowmoose Software/Alan Perry and
** placed in the public domain.
*/

#include <PalmOS.h>
#include "StopWatch.h"
#include "StopWatchRsrc.h"


/* Bitmaps and handles for the large digits, punctuation and labels */
static MemHandle        gBmHandle[bmCount];
static BitmapPtr        gBmPointer[bmCount];

static unsigned short   gTickWait;          /* Event loop timeout to        */
                                            /* update time display          */
static Boolean          gRunningClock;      /* Is the timer running now     */
static Boolean          gUpdateOff;         /* Don't update the clock       */
static Boolean          gDoReset;           /* Reset times at next start    */
static Boolean          gDisableAutoOff;    /* Is PalmOS AutoOff disabled   */
static Boolean          gFlashOn;           /* Flash On/Flash Off           */

static UInt32           gZeroTicks;         /* Tick count at start          */
static UInt32           gCumTicks;          /* Tick count before stop       */

/* Pre-calculated tick-to-time conversion factors */
static UInt32           gTicksPerHun;
static UInt32           gTicksPerSec;
static UInt32           gTicksPerMin;
static UInt32           gTicksPerHour;


/*
** Forward declarations
*/
static UInt32cvtTicksToHuns(    UInt32 ticks);
static UInt32 cvtTicksToSecs(   UInt32 ticks);
static UInt32 cvtTicksToMins(   UInt32 ticks);
static UInt32 cvtTicksToHours(  UInt32 ticks);

static void initBigDigitBms(    void);
static void displayBigSecs(     UInt32 seconds);
static void displayCurrTime(    UInt32 ticks);
static void displayAutoOffLabel(void);

static void swGetPrefs(         void);
static void swSetPrefs(         void);

static Boolean swFormEventHandler(EventPtr event);
static Boolean swHandleEvent(   EventPtr event);
static Err  swStartApplication(  void);
static void swStopApplication(   void);
static void swEventLoop(         void);


/*
** Inline functions to convert ticks to other time units
*/

static inline UInt32
cvtTicksToHuns(UInt32 ticks)
{
    return ((ticks / gTicksPerHun) % 100);
}


static inline UInt32
cvtTicksToSecs(UInt32 ticks)
{
    return ((ticks / gTicksPerSec) % 60);
}


static inline UInt32
cvtTicksToMins(UInt32 ticks)
{
    return ((ticks / gTicksPerMin) % 60);
}


static inline UInt32
cvtTicksToHours(UInt32 ticks)
{
    return ((ticks / gTicksPerHour) % 24);
}


/*
** Time display functions
*/

/*
** initBigDigitBms
** Initialize the Handles and Pointers for the bitmaps that contain the large
** format digits, punctuation and labels.
*/
static void
initBigDigitBms(void)
{
    int     i;

    for (i=0; i<bmCount; i++)
    {
        gBmHandle[i]    = DmGetResource('Tbmp', bmNumBase+i);
        gBmPointer[i]   = MemHandleLock(gBmHandle[i]);
    }

    return;
}


/*
** displayBigSecs
** Displays ticks in hour:minutes:seconds mode on one row and hundredths
** of a second on the next row.
*/
static void
displayBigSecs(UInt32 ticks)
{
    int h = cvtTicksToHours(ticks);
    int m = cvtTicksToMins(ticks);
    int s = cvtTicksToSecs(ticks);
    int c = cvtTicksToHuns(ticks);

    WinDrawBitmap(gBmPointer[h/10], 10, 37);
    WinDrawBitmap(gBmPointer[h%10], 33, 37);
    WinDrawBitmap(gBmPointer[bmColon-bmNumBase], 56, 37);
    WinDrawBitmap(gBmPointer[m/10], 62, 37);
    WinDrawBitmap(gBmPointer[m%10], 85, 37);
    WinDrawBitmap(gBmPointer[bmColon-bmNumBase], 108, 37);
    WinDrawBitmap(gBmPointer[s/10], 114, 37);
    WinDrawBitmap(gBmPointer[s%10], 137, 37);

    WinDrawBitmap(gBmPointer[c/10], 114, 78);
    WinDrawBitmap(gBmPointer[c%10], 137, 78);


    return;
}


/*
** displayCurrTime
** Displays the current time and its label.  The label for the field flashes if
** the clock is running.
*/
static void
displayCurrTime(UInt32 ticks)
{
    RectangleType   fmtLabelRect = { 0, 37, 10, 40 };

    if (!gUpdateOff)
    {
        displayBigSecs(ticks);

        if (gRunningClock)
        {
            /* if the clock is running, the "TIME" label flashes */
            if (gFlashOn)
            {
                WinDrawBitmap(gBmPointer[bmTime-bmNumBase], 0, 37);
                gFlashOn = false;
            }
            else
            {
                WinEraseRectangle(&fmtLabelRect, 0);
                gFlashOn = true;
            }
        }
        else
        {
            WinDrawBitmap(gBmPointer[bmTime-bmNumBase], 0, 37);
        }
    }

    return;
}


/*
** displayAutoOffLabel
** Displays (or blanks) the Auto-off label, depending on the global.
*/
static void
displayAutoOffLabel(void)
{
    RectangleType   aOffLabelRect = { kAutoOffX, kAutoOffY, 45, 10 };

    if (gDisableAutoOff)
    {
        WinEraseRectangle(&aOffLabelRect, 0);
    }
    else
    {
        WinDrawBitmap(gBmPointer[bmAutoOff-bmNumBase], kAutoOffX, kAutoOffY);
    }
    return;
}


/*
** Application preference functions
*/

/*
** swGetPrefs
** Gets the application preferences and sets the corresponding application
** globals appropriately.
*/
static void
swGetPrefs(void)
{
    Int16           rv;
    struct swPrefs  myPrefs;
    UInt16          myPrefsLen = sizeof(struct swPrefs);

    rv = PrefGetAppPreferences( kCreatorID, kBasicPrefID,
                                &myPrefs, &myPrefsLen, true);

    if (rv == noPreferenceFound)
    {
        gDisableAutoOff = true;
        swSetPrefs();
    }
    else
    {
        gDisableAutoOff = myPrefs.disableAutoOff;
    }

    return;
}


/*
** swSetPrefs
** Sets the application preferences based on the corresponding application
** globals
*/
static void
swSetPrefs(void)
{
    struct swPrefs  myPrefs;

    myPrefs.disableAutoOff  = gDisableAutoOff;

    PrefSetAppPreferences(  kCreatorID, kBasicPrefID, kPrefVersion,
                            &myPrefs, sizeof(struct swPrefs), true);

    return;
}


/*
** Application event functions
*/


/*
** swFormEventHandler
** event handler for theForm
*/
Boolean
swFormEventHandler(EventPtr event)
{
    UInt32  elapsedTime;
    FormPtr pForm;
    Boolean handled = false;

    switch (event->eType)
    {
        case kClockUpdateEvent:
            if (gRunningClock)
            {
                displayCurrTime((TimGetTicks() - gZeroTicks) + gCumTicks);
            }

            if (gDisableAutoOff)
                EvtResetAutoOffTimer();

            handled = true;
            break;

        case winEnterEvent:
            if ((event->data.winEnter.enterWindow ==
                (WinHandle)FrmGetFormPtr(theForm)) &&
                (event->data.winEnter.enterWindow ==
                (WinHandle)FrmGetFirstForm()))
            {
                gUpdateOff = false;
            }
            break;

        case winExitEvent:
            if (event->data.winExit.exitWindow ==
                (WinHandle)FrmGetFormPtr(theForm))
            {
                gUpdateOff = true;
            }
            break;

        case ctlSelectEvent:
            switch (event->data.ctlEnter.controlID)
            {
                case btStart:
                    if (!gRunningClock)
                    {
                        gRunningClock   = true;
                        gZeroTicks      = TimGetTicks();

                        if (gDoReset)
                        {
                            gCumTicks       = 0;
                            gDoReset        = false;
                        }

                        displayCurrTime(gCumTicks);
                    }

                    handled = true;
                    break;

                case btStop:
                    if (gRunningClock)
                    {
                        gRunningClock   = false;
                        elapsedTime     = TimGetTicks() - gZeroTicks;
                        gCumTicks       += elapsedTime;

                        displayCurrTime(gCumTicks);
                    }

                    handled = true;
                    break;

                case btReset:
                    if (gRunningClock)
                        gRunningClock = false;

                    displayCurrTime(0);

                    gDoReset    = true;
                    gCumTicks   = 0;

                    handled     = true;
                    break;
            }
            break;

        case frmOpenEvent:
            pForm  = FrmGetActiveForm();
            FrmDrawForm(pForm);

            if (FrmGetActiveFormID() == theForm)
            {
                displayCurrTime(0);
                displayAutoOffLabel();
            }

            handled         = true;
            break;

        case frmCloseEvent:
            pForm   = FrmGetActiveForm();
            FrmEraseForm(pForm);
            FrmDeleteForm(pForm);
            pForm   = 0;

            handled = true;
            break;

        case menuEvent:
            switch (event->data.menu.itemID)
            {
                case miAutoOff:
                    gDisableAutoOff = false;
                    displayAutoOffLabel();
                    swSetPrefs();

                    handled = true;
                    break;

                case miNoAutoOff:
                    gDisableAutoOff = true;
                    displayAutoOffLabel();
                    swSetPrefs();

                    handled = true;
                    break;

                case miAbout:
                    FrmAlert(aboutAlert);

                    handled = true;
                    break;

                case miHelp:
                    FrmHelp(helpStr);

                    handled = true;
                    break;
            }

            gUpdateOff = false;
            break;
    }

    return (handled);
} 


/*
** swHandleEvent
** Handle events for this application.
*/
static Boolean
swHandleEvent(EventPtr event) 
{
    int     formID;
    FormPtr pForm;
    Boolean handled = false;

    switch (event->eType) 
    {
        case frmLoadEvent:
            formID  = event->data.frmLoad.formID;
            pForm   = FrmInitForm(formID);
            FrmSetActiveForm(pForm);
            if (formID == theForm)
            {
                FrmSetEventHandler(pForm, swFormEventHandler);
            }

            handled = true;
            break;

        default:
            break;
    }

    return (handled);
}


/*
** swStartApplication
** Application initialization.
*/
static Err
swStartApplication()
{
    /* load the bitmaps for the big digits */
    initBigDigitBms();

    /* pre-compute these contants at run-time */
    gTicksPerHun    = SysTicksPerSecond() / 100L;
    gTicksPerSec    = SysTicksPerSecond();
    gTicksPerMin    = SysTicksPerSecond() * 60L;
    gTicksPerHour   = SysTicksPerSecond() * 3600L;

    gTickWait       = SysTicksPerSecond()/5;
    gRunningClock   = false;
    gUpdateOff      = false;
    gDoReset        = true;
    swGetPrefs();               /* sets gDisableAutoOff */

    gZeroTicks = gCumTicks = 0;

    FrmGotoForm(theForm);

    return (0);
}


/*
** swStopApplication
** Application clean-up.
*/
static void
swStopApplication()
{
    FrmCloseAllForms();
}


/*
** swEventLoop
** While the app is running, the core of this application.  Gets events
** and dispatches them.  Also, generates events to force clock display
** updates.
*/
static void
swEventLoop()
{
    EventType   event;
    UInt16      error;

    do
    {
        event.eType = nilEvent;

        EvtGetEvent(&event, gTickWait);

        if (event.eType == nilEvent)
        {
            event.eType = kClockUpdateEvent;
            EvtAddEventToQueue(&event);
        }
        else
        {
            if (SysHandleEvent(&event))
                continue;

            if (MenuHandleEvent(0, &event, &error))
                continue;

            if (swHandleEvent(&event))
                continue;

            FrmDispatchEvent(&event);
        }
    } while (event.eType != appStopEvent);

    return;
}


UInt32
PilotMain(
    UInt16      launchCode,
    void        *cmdPBP,
    UInt16      launchFlags)
{
    Err         err = 0;

    if (launchCode == sysAppLaunchCmdNormalLaunch)
    {
        if ((err = swStartApplication()) == 0)
        {
            swEventLoop();
            swStopApplication();
        }
    }

    return err;
}
